// -*- C++ -*-
//
// WeakPartonicDecayer.h is a part of Herwig++ - A multi-purpose Monte Carlo event generator
// Copyright (C) 2002-2007 The Herwig Collaboration
//
// Herwig++ is licenced under version 2 of the GPL, see COPYING for details.
// Please respect the MCnet academic guidelines, see GUIDELINES for details.
//
#ifndef HERWIG_WeakPartonicDecayer_H
#define HERWIG_WeakPartonicDecayer_H
//
// This is the declaration of the WeakPartonicDecayer class.
//

#include "PartonicDecayerBase.h"
#include "WeakPartonicDecayer.fh"

namespace Herwig {

using namespace ThePEG;

/**
 * The WeakPartonicDecayer class is designed to replace the HeavyDecayer
 * class which implements the partonic decays of hadrons containing a 
 * heavy quark in the same way as in FORTRAN Herwig. 
 *
 * There are a number of major changes
 *
 * - The helicity formalism is used for the decays so that the \f$\tau\f$ lepton
 *   gets the correct correlations (in reality the main effect is to ensure the
 *   \f$\tau\f$ is left-handed.
 *
 * - The particles produced directly by the hadronisation, i.e. the primary hadrons
 *   produced in cluster decay are checked to ensure that none of the exclusive
 *   modes are reproduced.
 *
 * - Two body modes are allowed to try and force baryon production etc. In this case
 *   the colours of the partons are connected.
 * 
 * - Three body modes of the form \f$q g \bar{q}\f$ are supported for penguin mediated
 *   weak decays.
 *
 *  Two types of matrix element are supported for this decay
 *
 *  - MECode=0   flat-phase space.
 *  - MECode=100 V-A matrix element for the heavy quark decay in the spectator model.
 *
 * @see HeavyDecayer
 */
class WeakPartonicDecayer: public PartonicDecayerBase {

public:

  /**
   * The default constructor.
   */
  inline WeakPartonicDecayer();

  /**
   * Check if this decayer can perfom the decay for a particular mode
   * @param parent The decaying particle
   * @param children The decay products
   * @return true If this decayer can handle the given mode, otherwise false.
   */
  virtual bool accept(tcPDPtr parent, const PDVector & children) const;

  
  /**
   *  Perform the decay of the particle to the specified decay products
   * @param parent The decaying particle
   * @param children The decay products
   * @return a ParticleVector containing the decay products.
   */
  virtual ParticleVector decay(const Particle & parent,
			       const PDVector & children) const;

  /**
   * Output the setup information for the particle database
   * @param os The stream to output the information to
   * @param header Whether or not to output the information for MySQL
   */
  virtual void dataBaseOutput(ofstream & os,bool header) const;

public:

  /** @name Functions used by the persistent I/O system. */
  //@{
  /**
   * Function used to write out object persistently.
   * @param os the persistent output stream written to.
   */
  void persistentOutput(PersistentOStream & os) const;

  /**
   * Function used to read in object persistently.
   * @param is the persistent input stream read from.
   * @param version the version number of the object when written.
   */
  void persistentInput(PersistentIStream & is, int version);
  //@}

  /**
   * The standard Init function used to initialize the interfaces.
   * Called exactly once for each class by the class description system
   * before the main function starts or
   * when this class is dynamically loaded.
   */
  static void Init();

public:

  /**
   * Weighting of phase space for V-A matrix elements
   */
  static double VAWt(Energy2, Energy2, Energy2, InvEnergy4);

protected:

  /** @name Clone Methods. */
  //@{
  /**
   * Make a simple clone of this object.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr clone() const;

  /** Make a clone of this object, possibly modifying the cloned object
   * to make it sane.
   * @return a pointer to the new object.
   */
  inline virtual IBPtr fullclone() const;
  //@}

private:

  /**
   * The static object used to initialize the description of this class.
   * Indicates that this is a concrete class with persistent data.
   */
  static ClassDescription<WeakPartonicDecayer> initWeakPartonicDecayer;

  /**
   * The assignment operator is private and must never be called.
   * In fact, it should not even be implemented.
   */
  WeakPartonicDecayer & operator=(const WeakPartonicDecayer &);

private:

  /**
   *  The code for the matrix element being used.
   */
  int MECode;
};

}

#include "ThePEG/Utilities/ClassTraits.h"

namespace ThePEG {

/** @cond TRAITSPECIALIZATIONS */

/** This template specialization informs ThePEG about the
 *  base classes of WeakPartonicDecayer. */
template <>
struct BaseClassTrait<Herwig::WeakPartonicDecayer,1> {
  /** Typedef of the first base class of WeakPartonicDecayer. */
  typedef Herwig::PartonicDecayerBase NthBase;
};

/** This template specialization informs ThePEG about the name of
 *  the WeakPartonicDecayer class and the shared object where it is defined. */
template <>
struct ClassTraits<Herwig::WeakPartonicDecayer>
  : public ClassTraitsBase<Herwig::WeakPartonicDecayer> {
  /** Return a platform-independent class name */
  static string className() { return "Herwig::WeakPartonicDecayer"; }
  /** Return the name of the shared library be loaded to get
   *  access to the WeakPartonicDecayer class and every other class it uses
   *  (except the base class). */
  static string library() { return "HwPartonicDecay.so"; }
};

/** @endcond */

}

#include "WeakPartonicDecayer.icc"

#endif /* HERWIG_WeakPartonicDecayer_H */
